function rotate(dir)
	local rot = {2,3,0,1}
	
	return rot[dir+1]	
end

function loadtile(x,y)
	local key = tostring(x) .. "," .. tostring(y)
	
	if (tiles[key] ~= nil) then
		return tiles[key]
	else
		return "error"
	end
end

function isthis(lookup,rule)
	for i,v in ipairs(lookup) do
		if (v[#v] == rule) then
			return true
		end
	end
	
	return false
end

function dynamictile(unit,x,y,name)
	local ox,oy = 0,0
	local result = 0
	local exclude = 0
	
	for i=1,4 do
		local v = ndirs[i]
		ox = v[1]
		oy = v[2]
		
		local tiles = findtype({name},x+ox,y+oy,unit)
		
		if (#tiles > 0) then
			result = result + 2 ^ (i - 1)
		end
	end
	
	return result
end

function dynamictiling()
	for i,unitid in ipairs(tiledunits) do
		local unit = mmf.newObject(unitid)
		unit.direction = dynamictile(unitid,unit.values[XPOS],unit.values[YPOS],unit.strings[UNITNAME])
	end
end

function dynamic(id)
	local unit = mmf.newObject(id)
	
	if (unit.values[TILING] == 1) then
		local x,y = unit.values[XPOS],unit.values[YPOS]
		local ox,oy = 0,0
		local name = unit.strings[UNITNAME]
		
		unit.direction = dynamictile(unit.fixed,x,y,name)
		
		for i=1,4 do
			local v = ndirs[i]
			ox = v[1]
			oy = v[2]
			
			local tiles = findtype({name},x+ox,y+oy,unit.fixed)
			
			if (#tiles > 0) then
				for a,b in ipairs(tiles) do
					local tile = mmf.newObject(b)

					tile.direction = dynamictile(b,x+ox,y+oy,name)
				end
			end
		end
	end
end

function dynamicat(x,y)
	local ox,oy = 0,0
	
	for i=1,4 do
		local v = ndirs[i]
		ox = v[1]
		oy = v[2]
		
		local tiles = findallhere(x+ox,y+oy)
		
		if (#tiles > 0) then
			for a,b in ipairs(tiles) do
				local tile = mmf.newObject(b)

				if (tile.values[TILING] == 1) then
					tile.direction = dynamictile(b,x+ox,y+oy,tile.strings[UNITNAME])
				end
			end
		end
	end
end

function colourset(id)
	local divider = string.find(id, "-")
	local name = string.sub(id, 1, divider - 1)
	local current = tonumber(string.sub(id, divider + 1))
	local oldcolour = 0
	if (colours[name] == nil) then
		colours[name] = current
	else
		oldcolour = colours[name]
		colours[name] = current
	end
	
	return oldcolour
end

function findall(name_)
	local result = {}
	local name = name_[1]
	
	for i,unit in ipairs(units) do
		if (unit.strings[UNITNAME] == name) or ((unit.strings[UNITTYPE] == "text") and (name == "text")) then
			if (name_[2] == nil) then
				table.insert(result, unit.fixed)
			else
				if ison(name_,unit.fixed) then
					table.insert(result, unit.fixed)
				end
			end
		end
	end
	
	return result
end

function pos(unitid)
	local unit = mmf.newObject(unitid)
	
	local x = unit.values[XPOS]
	local y = unit.values[YPOS]
	local dir = unit.values[DIR]
	local moved = unit.values[MOVED]
	
	return x,y,dir,moved
end

function delunit(unitid)
	local unit = mmf.newObject(unitid)
	local name = unit.strings[UNITNAME]
	local x,y = unit.values[XPOS],unit.values[YPOS]
	local tileid = x + y * roomsize
	local unitlist = unitlists[name]
	local unittype = unit.strings[UNITTYPE]
	
	if (unittype == "text") then
		updatecode = 1
	end
	
	for i,v in ipairs(unitlist) do
		if (v == unitid) then
			v = {}
			table.remove(unitlist, i)
		end
	end
	
	for i,v in ipairs(unitmap[tileid]) do
		if (v == unitid) then
			v = {}
			table.remove(unitmap[tileid], i)
		end
	end
	
	if (unit.strings[UNITTYPE] == "text") then
		for i,v in ipairs(codeunits) do
			if (v == unitid) then
				v = {}
				table.remove(codeunits, i)
			end
		end
	end
	
	if (unit.values[TILING] > 1) then
		for i,v in ipairs(animunits) do
			if (v == unitid) then
				v = {}
				table.remove(animunits, i)
			end
		end
	end
	
	if (unit.values[TILING] == 1) then
		for i,v in ipairs(tiledunits) do
			if (v == unitid) then
				v = {}
				table.remove(tiledunits, i)
			end
		end
	end
	
	for i,v in ipairs(units) do
		if (v.fixed == unitid) then
			v = {}
			table.remove(units, i)
		end
	end
end

function findtype(name_,x,y,this_)
	local result = {}
	local this = 0
	local tileid = x + y * roomsize
	local name = name_[1]
	
	if (this_ ~= nil) then
		this = this_
	end

	if (unitmap[tileid] ~= nil) then
		for i,v in ipairs(unitmap[tileid]) do
			if (v ~= this) then
				local unit = mmf.newObject(v)
				
				if (unit.strings[UNITNAME] == name) or ((unit.strings[UNITTYPE] == "text") and (name == "text")) then
					if (name_[2] == nil) then
						table.insert(result, v)
					else
						if ison(name_,v) then
							table.insert(result, v)
						end
					end
				end
			end
		end
	end
	
	return result
end

function findobstacle(x,y)
	local layer = map[0]
	local tile = layer:get_x(x,y)
	local result = {}
	local tileid = x + y * roomsize
	
	if (tile ~= 255) then
		table.insert(result, -1)
	end
	
	if (unitmap[tileid] ~= nil) then
		for i,v in ipairs(unitmap[tileid]) do
			table.insert(result, v)
		end
	end
	
	return result
end

function update(unitid,x,y,dir_)
	local unit = mmf.newObject(unitid)

	local unitname = unit.strings[UNITNAME]
	local dir,olddir = unit.values[DIR],unit.values[DIR]
	local tiling = unit.values[TILING]
	local unittype = unit.strings[UNITTYPE]
	local oldx,oldy = unit.values[XPOS],unit.values[YPOS]
	
	if (dir_ ~= nil) then
		dir = dir_
	end
	
	addundo({"update",unitname,oldx,oldy,olddir,x,y,dir,unit.values[ID]})
	
	unit.values[XPOS] = x
	unit.values[YPOS] = y
	unit.values[DIR] = dir
	unit.values[MOVED] = 1
	unit.values[POSITIONING] = 0
	
	updateunitmap(unitid,oldx,oldy,x,y)
	
	if (tiling == 1) then
		dynamic(unitid)
		dynamicat(oldx,oldy)
	end
	
	if (unittype == "text") then
		updatecode = 1
	end
end

function updatedir(unitid,dir)
	local unit = mmf.newObject(unitid)
	local x,y = unit.values[XPOS],unit.values[YPOS]
	local unitname = unit.strings[UNITNAME]
	local unittype = unit.strings[UNITTYPE]
	local olddir = unit.values[DIR]
	
	addundo({"update",unitname,x,y,olddir,x,y,dir,unit.values[ID]})
	unit.values[DIR] = dir
	
	if (unittype == "text") then
		updatecode = 1
	end
end

function findtext(x,y)
	local tileid = x + y * roomsize
	
	if (unitmap[tileid] ~= nil) then
		for i,v in ipairs(unitmap[tileid]) do
			local unit = mmf.newObject(v)
			
			if (unit.strings[UNITTYPE] == "text") then
				return v
			end
		end
	end
	
	return 0
end

function material(name)
	local protection = hasfeature(name,"is",name)
	
	if (protection == nil) then
		if (unitlists[name] ~= nil) then
			return 1
		end
	end
	
	return 0
end

function findallhere(x,y,exclude_)
	local result = {}
	
	local exclude = 0
	
	if (exclude_ ~= nil) then
		exclude = exclude_
	end
	
	local tileid = x + y * roomsize
	
	if (unitmap[tileid] ~= nil) then
		for i,unitid in ipairs(unitmap[tileid]) do
			if (unitid ~= exclude) then
				local unit = mmf.newObject(unitid)
				if (unit.values[XPOS] == x) and (unit.values[YPOS] == y) then
					table.insert(result, unitid)
				end
			end
		end
	end
	
	return result
end

function findempty()
	local result = {}
	local array = {}
	local layer = map[0]
	
	for i,unit in ipairs(units) do
		local x,y = unit.values[XPOS],unit.values[YPOS]
		local arrayid = x + y * roomsize
		
		if (array[arrayid] == nil) then
			array[arrayid] = {}
			table.insert(array[arrayid], unit.fixed)
		end
	end
	
	for i=0,roomsize-1 do
		for j=0,roomsize-1 do
			local empty = 1
			local tile = layer:get_x(i,j)
			local arrayid = i + j * roomsize
			
			if (tile ~= 255) or (array[arrayid] ~= nil) then
				empty = 0
			end
			
			if (empty == 1) then
				table.insert(result, i + j * roomsize)
			end
		end
	end	
	
	return result
end

function delete(unitid,shake)
	if (deleted[unitid] == nil) then
		local unit = mmf.newObject(unitid)
		
		local x,y,dir = unit.values[XPOS],unit.values[YPOS],unit.values[DIR]
		local unitname = unit.strings[UNITNAME]
		local insidename = unit.strings[UNITNAME]
		
		if (unit.strings[UNITTYPE] == "text") then
			insidename = "text"
		end
		
		inside(unitname,x,y,dir,unitid)
		
		addundo({"remove",unitname,x,y,dir,unit.values[ID]})
		unit = {}
		delunit(unitid)
		MF_remove(unitid,shake)
		
		dynamicat(x,y)
		
		deleted[unitid] = 1
	else
		print("already deleted")
	end
end

function findwalls(x,y)
	local result = {}
	local stop = findfeature(nil,"is","stop")
	local push = findfeature(nil,"is","push")
	local pull = findfeature(nil,"is","pull")
	local layer = map[0]
	
	if (stop ~= nil) then
		for i,v in ipairs(stop) do
			local stops = findtype(v,x,y,0)
			
			if (#stops > 0) then
				for a,b in ipairs(stops) do
					table.insert(result, b)
				end
			end
		end
	end
	
	if (push ~= nil) then
		for i,v in ipairs(push) do
			local pushes = findtype(v,x,y,0)
			
			if (#pushes > 0) then
				for a,b in ipairs(pushes) do
					table.insert(result, b)
				end
			end
		end
	end
	
	if (pull ~= nil) then
		for i,v in ipairs(pull) do
			local pulls = findtype(v,x,y,0)
			
			if (#pulls > 0) then
				for a,b in ipairs(pulls) do
					table.insert(result, b)
				end
			end
		end
	end
	
	if (layer:get_x(x,y) ~= 255) then
		table.insert(result, 1)
	end
	
	return result
end

function listrules()
	for i,v in ipairs(features) do
		local text = ""
		
		for a,b in ipairs(v) do
			if (type(b) == "string") then
				text = text .. b .. " "
			elseif (type(b) == "table") then
				for c,d in ipairs(b) do
					text = text .. d .. " "
					
					if (c ~= #b) then
						text = text .. "and "
					end
				end
			end
		end
		
		print(text)
	end
end

function findfeatureat(rule1,rule2,rule3,x,y)
	local result = {}
	local targets = findfeature(rule1,rule2,rule3)
	
	if (targets ~= nil) then
		for i,v in ipairs(targets) do
			local name = v[1]
			for a,unit in ipairs(units) do
				local unitx,unity = unit.values[XPOS],unit.values[YPOS]
				if (unit.values[XPOS] == x) and (unit.values[YPOS] == y) then
					if (unit.strings[UNITNAME] == name) or ((unit.strings[UNITTYPE] == "text") and (name == "text")) then
						if (v[2] == nil) then
							table.insert(result, unit.fixed)
						else
							if ison(v,unit.fixed) then
								table.insert(result, unit.fixed)
							end
						end
					end
				end
			end
		end
	end
	
	if (#result > 0) then
		return result
	else
		return nil
	end
end

function convert(stuff,mats)
	local layer = map[0]
	local delthese = {}
	local mat1 = stuff
	
	if (mat1 ~= "empty") then
		if (unitlists[mat1] ~= nil) then
			for i,unitid in ipairs(unitlists[mat1]) do
				local unit = mmf.newObject(unitid)
				local x,y,dir,id,name = unit.values[XPOS],unit.values[YPOS],unit.values[DIR],unit.values[ID],unit.strings[UNITNAME]
				local delthis = false
				
				for a,matdata in ipairs(mats) do
					local mat2 = matdata[1]
					
					if (mat2 ~= "empty") then
						if (matdata[2] == nil) or ((matdata[2] ~= nil) and (ison({mat1,matdata[2]},unit.fixed))) then
							local newunitid = MF_emptycreate(mat2,x,y)
							local newunit = mmf.newObject(newunitid)
							
							newunit.values[ONLINE] = 1
							newunit.values[XPOS] = x
							newunit.values[YPOS] = y
							newunit.values[DIR] = dir
							newunit.values[ID] = id
							newunit.values[EFFECT] = 1
							newunit.flags[9] = true
							
							addunit(newunitid)
							addunitmap(newunitid,x,y)
							dynamic(newunitid)
							
							addundo({"create",newunit.strings[UNITNAME],id})
							delthis = true
							
							if (newunit.strings[UNITTYPE] == "text") then
								updatecode = 1
							end
						end
					else
						if (matdata[2] == nil) or ((matdata[2] ~= nil) and (ison({mat1,matdata[2]},unit.fixed))) then
							updateunitmap(unitid,x,y,x,y)
							delthis = true
						end
					end
				end
				
				if delthis then
					table.insert(delthese, unitid)
					if (unit.strings[UNITTYPE] == "text") then
						updatecode = 1
						print("text deleted")
					end
				end
			end
		end
	else
		for a,matdata in ipairs(mats) do
			local mat2 = matdata[1]
			
			if (matdata[2] == nil) then
				if (mat2 ~= "empty") then
					local convunitmap = {}
					
					for a,unit in ipairs(units) do
						local tileid = unit.values[XPOS] + unit.values[YPOS] * roomsize
						convunitmap[tileid] = 1
					end
					
					for i=0,roomsize-1 do
						for j=0,roomsize-1 do
							local empty = 1
							
							local tileid = i + j * roomsize
							if (convunitmap[tileid] ~= nil) then
								empty = 0
							end
							
							if (layer:get_x(i,j) ~= 255) then
								empty = 0
							end
							
							if (empty == 1) then
								local newunitid = MF_emptycreate(mat2,i,j)
								local newunit = mmf.newObject(newunitid)
								
								local id = newid()
								local dir = math.random(0,3)
								
								newunit.values[ONLINE] = 1
								newunit.values[XPOS] = i
								newunit.values[YPOS] = j
								newunit.values[DIR] = dir
								newunit.values[ID] = id
								newunit.values[EFFECT] = 1
								newunit.flags[9] = true
								
								addunit(newunitid)
								addunitmap(newunitid,i,j)
								dynamic(newunitid)
								
								addundo({"create",newunit.strings[UNITNAME],id})
								
								if (newunit.strings[UNITTYPE] == "text") then
									updatecode = 1
								end
							end
						end
					end
				end
			end
		end
	end
	
	if (#delthese > 0) then
		for i,unitid in ipairs(delthese) do
			local unit = mmf.newObject(unitid)
			addundo({"remove",unit.strings[UNITNAME],unit.values[XPOS],unit.values[YPOS],unit.values[DIR],unit.values[ID]})
			delunit(unitid)
			MF_specialremove(unitid,0,2)
		end
	end
end

function copy(unitid,x,y)
	for i,unit in ipairs(units) do
		if (unit.fixed == unitid) then
			local oldx,oldy,dir,name = unit.values[XPOS],unit.values[YPOS],unit.values[DIR],unit.strings[UNITNAME]
			
			create(name,x,y,dir,oldx,oldy)
		end
	end
end

function create(name,x,y,dir,oldx_,oldy_)
	local oldx,oldy = x,y
	local tileid = x + y * roomsize
	
	if (oldx_ ~= nil) then
		oldx = oldx_
	end
	
	if (oldy_ ~= nil) then
		oldy = oldy_
	end
	
	local newunitid = MF_emptycreate(name,oldx,oldy)
	local newunit = mmf.newObject(newunitid)
	
	local id = newid()
	
	newunit.values[ONLINE] = 1
	newunit.values[XPOS] = x
	newunit.values[YPOS] = y
	newunit.values[DIR] = dir
	newunit.values[ID] = id
	
	newunit.flags[9] = true
	addunit(newunitid)
	addunitmap(newunitid,x,y)
	dynamic(newunitid)
	
	addundo({"create",name,id})
end

function getunitid(id)
	for i,unit in ipairs(units) do
		if (unit.values[ID] == id) then
			return unit.fixed
		end
	end
	
	return 0
end

function newid()
	local result = currid
	currid = currid + 1
	return result
end

function addunitmap(id,x,y)
	local tileid = x + y * roomsize
	
	if (unitmap[tileid] == nil) then
		unitmap[tileid] = {}
	end
	
	table.insert(unitmap[tileid], id)
end

function updateunitmap(id,oldx,oldy,x,y)
	local tileid = x + y * roomsize
	local oldtileid = oldx + oldy * roomsize
	
	if (unitmap[oldtileid] ~= nil) then
		for i,v in ipairs(unitmap[oldtileid]) do
			if (v == id) then
				table.remove(unitmap[oldtileid], i)
			end
		end
	end
	
	addunitmap(id,x,y)
end

function inside(name,x,y,dir,unitid)
	local ins = findfeature(nil,"in",name)
	
	if (ins ~= nil) then
		for i,v in ipairs(ins) do
			if (v[2] == nil) or ((v[2] ~= nil) and ison(v,unitid)) then
				if (v[1] ~= "text") then
					for a,b in ipairs(materials) do
						if (b == v[1]) and (v[1] ~= "empty") then
							if (v[1] ~= "all") then
								create(v[1],x,y,dir)
							else
								createall(v,x,y,unitid)
							end
						end
					end
				else
					create("text_" .. name,x,y,dir)
				end
			end
		end
	end
end

function ison(findresult,unitid)
	local result = false
	local unit = mmf.newObject(unitid)
	local x,y = unit.values[XPOS],unit.values[YPOS]
	local id = x + y * roomsize
	local checked = {}
	
	if (findresult[2] ~= nil) then
		local required = #findresult[2]
		local found = 0
		if (unitmap[id] ~= nil) then
			for i,v in ipairs(unitmap[id]) do
				local unithere = mmf.newObject(v)
				
				for a,b in ipairs(findresult[2]) do
					if (unithere.strings[UNITNAME] == b) and (checked[b] == nil) then
						found = found + 1
						checked[b] = 1
					end
				end
				
				if (found == required) then
					result = true
				end
			end
		end
	else
		result = true
	end
	
	return result
end

function ison_full(found,unitid)
	for i,v in ipairs(found) do
		if (v[2] == nil) then
			return true
		else
			return ison(v,unitid)
		end
	end
end

function hasfeature(rule1,rule2,rule3,unitid)
	if (featureindex[rule1] ~= nil) then
		local v1,v2,v3,v4 = "","","",nil
		for i,v in ipairs(featureindex[rule1]) do
			if (#v == 3) then
				v1 = v[1]
				v2 = v[2]
				v3 = v[3][1]
				v4 = nil
			elseif (#v == 5) then
				v1 = v[1]
				v2 = v[4]
				v3 = v[5]
				v4 = v[3]
			end
			
			if (v1 == rule1) and (v2 == rule2) and (v3 == rule3) then
				if (v4 == nil) then
					return 1
				else
					if (rule1 ~= "empty") then
						if ison({v1,v4},unitid) then
							return 1
						end
					end
				end
			end
		end
	end
	
	if (featureindex[rule3] ~= nil) then
		local v1,v2,v3,v4 = "","","",nil
		for i,v in ipairs(featureindex[rule3]) do
			if (#v == 3) then
				v1 = v[1]
				v2 = v[2]
				v3 = v[3][1]
				v4 = nil
			elseif (#v == 5) then
				v1 = v[1]
				v2 = v[4]
				v3 = v[5]
				v4 = v[3]
			end
			
			if (v1 == rule1) and (v2 == rule2) and (v3 == rule3) then
				if (v4 == nil) then
					return 1
				else
					if (rule1 ~= "empty") then
						local test = mmf.newObject(unitid)
						if ison({v1,v4},unitid) then
							return 1
						end
					end
				end
			end
		end
	end
	
	return nil
end

function animate()
	for i,unitid in ipairs(animunits) do
		local unit = mmf.newObject(unitid)
		
		if (unit.values[TILING] == 3) then
			unit.values[VISUALDIR] = (unit.values[VISUALDIR] + 1) % 2
			unit.direction = unit.values[DIR] * 8 + unit.values[VISUALDIR]
		end
		
		if (unit.values[TILING] == 2) then
			unit.direction = unit.values[DIR] * 8 + unit.values[VISUALDIR]
		end
	end
end